
# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'nonassocARROWnonassocCOMMAleftASSIGNleftRBRACKETLBRACKETleftKEYWORDNAMELITERALrightMODIF_QMARKMODIF_STARMODIF_PLUSARROW ASSIGN COMMA KEYWORD LBRACKET LITERAL MODIF_PLUS MODIF_QMARK MODIF_STAR NAME RBRACKET\n        DOCUMENT : MACRO_CHAIN\n                 | VARIABLE\n        \n        DOCUMENT : DOCUMENT MACRO_CHAIN\n                 | DOCUMENT VARIABLE\n         MACRO_CHAIN : MACRO ARROW MACRO \n        MACRO : MACRO MODIF_PLUS\n              | MACRO MODIF_STAR\n              | MACRO MODIF_QMARK\n         MACRO : KEYWORD  MACRO : KEYWORD LBRACKET ARGS RBRACKET  VARIABLE_NAME : NAME  VARIABLE : NAME ASSIGN ARG  ARGS : ARGS COMMA ARG  ARGS : ARG  ARG : LITERAL  ARG : MACRO  ARG : VARIABLE_NAME '
    
_lr_action_items = {'NAME':([0,1,2,3,6,7,8,10,11,12,13,14,15,16,17,18,19,20,23,24,],[5,5,-1,-2,-9,-3,-4,-6,-7,-8,16,16,-5,-11,-12,-15,-16,-17,-10,16,]),'KEYWORD':([0,1,2,3,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,23,24,],[6,6,-1,-2,-9,-3,-4,6,-6,-7,-8,6,6,-5,-11,-12,-15,-16,-17,-10,6,]),'$end':([1,2,3,6,7,8,10,11,12,15,16,17,18,19,20,23,],[0,-1,-2,-9,-3,-4,-6,-7,-8,-5,-11,-12,-15,-16,-17,-10,]),'ARROW':([4,6,10,11,12,23,],[9,-9,-6,-7,-8,-10,]),'MODIF_PLUS':([4,6,10,11,12,15,19,23,],[10,-9,-6,-7,-8,10,10,-10,]),'MODIF_STAR':([4,6,10,11,12,15,19,23,],[11,-9,-6,-7,-8,11,11,-10,]),'MODIF_QMARK':([4,6,10,11,12,15,19,23,],[12,-9,-6,-7,-8,12,12,-10,]),'ASSIGN':([5,],[13,]),'RBRACKET':([6,10,11,12,16,18,19,20,21,22,23,25,],[-9,-6,-7,-8,-11,-15,-16,-17,23,-14,-10,-13,]),'COMMA':([6,10,11,12,16,18,19,20,21,22,23,25,],[-9,-6,-7,-8,-11,-15,-16,-17,24,-14,-10,-13,]),'LBRACKET':([6,],[14,]),'LITERAL':([13,14,24,],[18,18,18,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'DOCUMENT':([0,],[1,]),'MACRO_CHAIN':([0,1,],[2,7,]),'VARIABLE':([0,1,],[3,8,]),'MACRO':([0,1,9,13,14,24,],[4,4,15,19,19,19,]),'ARG':([13,14,24,],[17,22,25,]),'VARIABLE_NAME':([13,14,24,],[20,20,20,]),'ARGS':([14,],[21,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> DOCUMENT","S'",1,None,None,None),
  ('DOCUMENT -> MACRO_CHAIN','DOCUMENT',1,'p_document','parser.py',38),
  ('DOCUMENT -> VARIABLE','DOCUMENT',1,'p_document','parser.py',39),
  ('DOCUMENT -> DOCUMENT MACRO_CHAIN','DOCUMENT',2,'p_document_list','parser.py',46),
  ('DOCUMENT -> DOCUMENT VARIABLE','DOCUMENT',2,'p_document_list','parser.py',47),
  ('MACRO_CHAIN -> MACRO ARROW MACRO','MACRO_CHAIN',3,'p_macro_chain','parser.py',53),
  ('MACRO -> MACRO MODIF_PLUS','MACRO',2,'p_macro_w_modif','parser.py',59),
  ('MACRO -> MACRO MODIF_STAR','MACRO',2,'p_macro_w_modif','parser.py',60),
  ('MACRO -> MACRO MODIF_QMARK','MACRO',2,'p_macro_w_modif','parser.py',61),
  ('MACRO -> KEYWORD','MACRO',1,'p_macro_wo_args','parser.py',68),
  ('MACRO -> KEYWORD LBRACKET ARGS RBRACKET','MACRO',4,'p_macro_w_args','parser.py',74),
  ('VARIABLE_NAME -> NAME','VARIABLE_NAME',1,'p_variable','parser.py',80),
  ('VARIABLE -> NAME ASSIGN ARG','VARIABLE',3,'p_variable_from_arg','parser.py',84),
  ('ARGS -> ARGS COMMA ARG','ARGS',3,'p_arg_list','parser.py',88),
  ('ARGS -> ARG','ARGS',1,'p_args','parser.py',92),
  ('ARG -> LITERAL','ARG',1,'p_arg','parser.py',96),
  ('ARG -> MACRO','ARG',1,'p_arg_from_macro','parser.py',100),
  ('ARG -> VARIABLE_NAME','ARG',1,'p_arg_from_var','parser.py',104),
]



import logging

import rita

from functools import reduce
from itertools import zip_longest

from rita.utils import Node

logger = logging.getLogger(__name__)


def add_implicit_punct(rules, config):
    """
    When writing rule, user usually doesn't care about some punct characters between words.
    We add them implicitly (unless this setting is turned off)
    """
    for group_label, pattern in rules:
        def gen():
            for p in pattern:
                yield p
                yield ("punct", None, "?")

        if len(pattern) == 1:
            yield (group_label, pattern)
        else:
            yield (group_label, list(gen())[:-1])


def handle_multi_word(rules, config):
    """
    spaCy splits everything in tokens. Words with dash ends up in different tokens.
    We don't want for user to even care about this, so we make this work implicitly
    
    WORD("Knee-length") => WORD("Knee"), WORD("-"), WORD("length")
    """
    for group_label, pattern in rules:
        yield (group_label, pattern)


def is_complex(arg):
    splitters = ["-"]
    return any([s in arg
                for s in splitters])

def has_complex(args):
    """
    Tells if any of arguments will be impacted by tokenizer
    """
    
    return any([is_complex(a)
                for a in args])


def branch_pattern(pattern, config):
    """
    Creates multiple lists for each possible permutation
    """
    root = Node()
    current = root
    depth = 0
    for idx, p in enumerate(pattern):
        if p[0] == "either":
            n = Node()
            current.add_next(n)
            current = n
            current.depth = depth
            for e in p[1]:
                current.add_child(e(config=config))
                depth += 1
        elif p[0] == "any_of" and has_complex(p[1]):
            _all = set(p[1])
            _complex = set(filter(is_complex, _all))
            simple = _all - _complex
            n = Node()
            current.add_next(n)
            current = n
            current.depth = depth
            current.add_child(("any_of", simple, p[2]))
            for c in sorted(_complex):
                current.add_child(("phrase", c, p[2]))
                depth += 1
        else:
            n = Node(p)
            current.add_next(n)
            current = n
            current.depth = depth

    for p in root.unwrap():
        yield p


def handle_rule_branching(rules, config):
    """
    If we have an OR statement, eg. `WORD(w1)|WORD(w2)`,
    Generic approach is to clone rules and use w1 in one, w2 in other.
    It may be an overkill, but some situations are not covered in simple approach
    """ 
    for group_label, pattern in rules:
        # Covering WORD(w1)|WORD(w2) case
        if any([p == "either"
                for (p, _, _) in pattern]):
            for p in branch_pattern(pattern, config):
                yield (group_label, p)

        # Covering case when there are complex items in list
        elif any([p == "any_of" and has_complex(o)
                  for (p, o, _) in pattern]):
            for p in branch_pattern(pattern, config):
                yield (group_label, p)
        else:
            yield (group_label, pattern)


def dummy(rules, config):
    """
    Placeholder which does nothing
    """
    logger.debug("Initial rules: {}".format(rules))
    return rules


def rule_tuple(d):
    return (d["label"], d["data"])


def preprocess_rules(root, config):
    logger.info("Preprocessing rules")

    rules = [rule_tuple(doc())
             for doc in root
             if doc and doc()]

    pipeline = [dummy, handle_rule_branching, handle_multi_word]

    if config.implicit_punct:
        logger.info("Adding implicit Punctuations")
        pipeline.append(add_implicit_punct)

    return reduce(lambda acc, p: p(acc, config), pipeline, rules)



import logging
import re

from functools import partial

logger = logging.getLogger(__name__)

def apply_operator(syntax, op):
    if not op:
        return syntax

    elif op == "!": # A bit complicated one
        return ("((?!{})\w+)".format(syntax
                            .rstrip(")")
                            .lstrip("(")))
    else:
        return syntax + op


def any_of_parse(lst, op=None):
    return apply_operator(r"({0})".format("|".join(sorted(lst))), op)


def regex_parse(r, op=None):
    return apply_operator(r, op)


def not_supported(key, *args, **kwargs):
    raise RuntimeError(
        "Rule '{0}' is not supported in standalone mode"
        .format(key)
    )


def person_parse(op=None):
    return apply_operator(r"([A-Z]\w+\s?)", op)


def entity_parse(value, op=None):
    if value == "PERSON":
        return person_parse(op)
    else:
        return not_supported(value)


def punct_parse(_, op=None):
    return apply_operator(r"[.,!;?:]", op)


def word_parse(value, op=None):
    if value:
        initial = r"({})".format(value)
    else:
        initial = r"(\w+)"

    return apply_operator(initial, op)

def fuzzy_parse(r, op=None):
    # TODO: build premutations
    return apply_operator(r"({0})[.,?;!]?".format("|".join(r)), op)


def whitespace_parse(_, op=None):
    return r"\s"

def phrase_parse(value, op=None):
    return apply_operator(r"({})".format(value), op)


PARSERS = {
    "any_of": any_of_parse,
    "value": word_parse,
    "regex": regex_parse,
    "entity": entity_parse,
    "lemma": partial(not_supported, "LEMMA"),
    "pos": partial(not_supported, "POS"),
    "punct": punct_parse,
    "fuzzy": fuzzy_parse,
    "whitespace": whitespace_parse,
    "phrase": phrase_parse,
}


def rules_to_patterns(label, data):
    logger.debug("data: {}".format(data))
    def gen():
        """
        Implicitly add spaces between rules
        """
        if len(data) == 0:
            return
        
        yield data[0]
        for (t, d, op) in data[1:]:
            if t != "punct":
                yield ("whitespace", None, None) 
            yield (t, d, op)
    
    return (
        label,
        [PARSERS[t](d, op) for (t, d, op) in gen()],
    )


class RuleExecutor(object):
    def __init__(self, patterns, config):
        self.config = config
        self.patterns = [self.compile(label, rules)
                         for label, rules in patterns]

    def compile(self, label, rules):
        flags = re.DOTALL
        if self.config.ignore_case:
            flags = flags | re.IGNORECASE
            
        return re.compile(r"(?P<{0}>{1})".format(label, "".join(rules)), flags)

    def execute(self, text):
        for p in self.patterns:
            for match in p.finditer(text):
                yield {
                    "start": match.start(),
                    "end": match.end(),
                    "text": match.group(),
                    "label": match.lastgroup,
                }


def compile_rules(rules, config):
    logger.info("Using standalone rule implementation")
    patterns = [rules_to_patterns(*group) for group in rules]
    executor = RuleExecutor(patterns, config)
    return executor
